import os

from osgeo import gdal
import rioxarray as rxr
import xarray as xr
from enum import Enum

from geopy.distance import distance
import numpy as np
import pandas as pd
import datetime


def gdal_corners(filename):
    """
    Extracts the corner coordinates (bounding box) of a raster image using GDAL.

    Parameters:
    -----------
    filename : str
        The file path to the raster image.

    Returns:
    --------
    tuple
        A tuple containing the minimum and maximum x and y coordinates of the raster image:
        (minx, miny, maxx, maxy).
        - minx: Minimum x coordinate (longitude).
        - miny: Minimum y coordinate (latitude).
        - maxx: Maximum x coordinate (longitude).
        - maxy: Maximum y coordinate (latitude).

    Notes:
    ------
    The function assumes the image has a georeferenced coordinate system and uses the
    GeoTransform matrix to compute the corner coordinates.

    Example:
    --------
    >>> minx, miny, maxx, maxy = gdal_corners('image.tif')
    >>> print(f"Bounding box: {minx}, {miny}, {maxx}, {maxy}")
    """
    ds = gdal.Open(filename)
    width = ds.RasterXSize
    height = ds.RasterYSize
    gt = ds.GetGeoTransform()
    minx = gt[0]
    miny = gt[3] + width * gt[4] + height * gt[5]
    maxx = gt[0] + width * gt[1] + height * gt[2]
    maxy = gt[3]
    return minx, miny, maxx, maxy


def join_files(file_list, version = "v1"):
    """
        a method for merging raster/tif data along the band dimension using
        rioxarray. this method does not save data to storage.
    """

    # initialize band names
    var_names = []

    # loop over bands, append to a rasterio xarray
    for file in file_list:

        xmin, ymin, xmax, ymax = tuple(gdal_corners(file))
        cda = rxr.open_rasterio(file, chunks=(1, 1200, 1200))
        cda = cda.sel(x=slice(xmin, xmax), y=slice(ymax, ymin))

        if file == file_list[0]:
            data = cda
        else:
            data = xr.concat([data, cda], "band")

        # extract frequency / polarization band
        if version == "v1":
            file_name = file.split('/')[-1]
            var_name = file_name.split('_')[5]
            var_name = var_name[0:2] + var_name[5:]
            var_names.append(var_name)
        elif version == "v3":
            file_name = file.split('/')[-1]
            var_name = file_name.split('_')[7]
            var_names.append(var_name)


    # change band names for reading
    data = data.assign_coords({'band': var_names})
    return data


def join_sar_radiom(da, radiom):
    """
    input
        da - rioxarray containing SWESARR SAR data. 6 channels expected.
        radiom - pandas array containing radiometer data. 3 channels expected.

    output
        data_p - pandas data series intended for plotting with hvplot's "groupby" feature.
                sadly, all measurement data is crammed into a single column, making this
                difficult to use outside of hvplot functionality.
        out_data - pandas data series that is readable. intended for use in student exercise.

    TO-DO / IMPROVEMENTS :
        implementing this with xarray would be much more RAM efficient.

    """
    # first, convert the data from the SAR's meter-based,
    # universal transverse mercator (UTM) coordinate system
    # to the radiometer's old-fashioned
    # latitude/longitude coordinate system
    sar_geo = da.rio.reproject("EPSG:4326")

    # get latidue and longitude from SAR data
    lat_sar = sar_geo.y.data
    lon_sar = sar_geo.x.data

    # radiometer latitude/longitude values as numpy arrays
    lat_rad = radiom['Latitude (deg)'].to_numpy()
    lon_rad = radiom['Longitude (deg)'].to_numpy()

    # loop over latitude and longitude
    frames = []
    k = -1
    for lat, lon in zip(lat_rad, lon_rad):
        k += 1

        # get the difference between latitude and longitude pairs
        # for SAR and radiometer data
        lat_m = np.abs(lat_sar - lat).tolist()
        lon_m = np.abs(lon_sar - lon).tolist()

        # use python's built-in functions to find minimum index
        ind_lat = lat_m.index(min(lat_m))
        ind_lon = lon_m.index(min(lon_m))

        # write sar lat and lon
        s_lat = lat_sar[ind_lat]
        s_lon = lon_sar[ind_lon]

        # get distance between the estimated sar and radiometer center positions
        # using vincenty's formula from the geopy library
        dis = distance((lat, lon), (s_lat, s_lon)).m

        # construct data dictionary
        data_d = {'sar_lat': s_lat, 'sar_lon': s_lon,
                  'rad_lat': lat, 'rad_lon': lon,
                  'ind_lat': ind_lat, 'ind_lon': ind_lon,
                  'dist_m': dis}

        # make a pandas dataframe based on the above data!
        df = pd.DataFrame(data_d, index=[k])

        # throw it in an array for good measure
        frames.append(df)

    # use "list comprehension" syntax to merge the pandas dataframes together
    location_data = pd.concat(data for data in frames)
    del frames, df, data_d

    # now lets store our SAR data based on our filtered results
    sar_data = []

    for in1, in2 in zip(location_data['sar_lon'].tolist(), location_data['sar_lat'].tolist()):
        # access the dask array storing our results
        d = sar_geo.sel(x=in1, y=in2).compute().data.tolist()
        # append to our list
        sar_data.append(d)

    # convert both arrays array to a numpy array for easy merging
    data = np.array(sar_data)
    radiom_d = radiom.iloc[:, 4:7].to_numpy()

    # insert the radiometer data to the SAR data as a column vector
    for i in range(np.size(radiom_d, 1)):
        data = np.insert(data, np.size(data, 1), radiom_d[:, i], axis=1)
    del radiom_d

    # the following section is for plotting with hvplot while including a label only.
    #
    # use list operations paired with pandas series to repeat the lat/lon data
    lon_ser = pd.Series(location_data['sar_lon'].to_list() * (6) \
                        + location_data['rad_lon'].to_list() * (3))
    lat_ser = pd.Series(location_data['sar_lat'].to_list() * (6) \
                        + location_data['rad_lat'].to_list() * (3))

    # flatten MATLAB/*F*ortran style. (Default flattens as row vectors)
    data_ser = data.flatten(order='F')

    # get series length, create IDs for plotting
    # all series have the same length, so this shouldn't matter.
    sl = len(radiom['TB X (K)'])
    id_ser = pd.Series(
        ['09VV SAR'] * sl + ['09VH SAR'] * sl + ['13VV SAR'] * sl + ['13VH SAR'] * sl + \
        ['17VV SAR'] * sl + ['17VH SAR'] * sl + \
        ['X-band Rad'] * sl + ['K-band Rad'] * sl + ['Ka-band Rad'] * sl, name="ID"
    )

    # the final frame used only for plotting by groups with hvplot
    frame = {'Longitude (deg)': lon_ser, 'Latitude (deg)': lat_ser,
             'Measurements': data_ser, 'ID': id_ser}
    data_p = pd.DataFrame(frame)

    # convert swesarr numpy data to dataframe
    swesarr_df = pd.DataFrame(data=data,
                              columns=["09VV SAR", "09VH SAR",
                                       "13VV SAR", "13VH SAR",
                                       "17VV SAR", "17VH SAR",
                                       'X-band Rad', 'Ku-band Rad',
                                       'Ka-band Rad'])

    # Convert UTC string to date time array
    times = radiom.UTC.to_frame()
    times = times['UTC'].to_list()

    # convert to datetime
    times = [datetime.datetime.strptime(time_str, '%Y%m%d-%H:%M:%S.%f') for time_str in times]

    radiom.UTC = times

    # combine time data with swesarr measurements
    out_data = pd.concat([radiom.UTC.to_frame(), swesarr_df], axis=1)

    # return the variable used for plotting and its more user-friendly variant.
    return data_p, out_data


if __name__ == '__main__':
    print('name is main!')
    print(os.listdir())


import datetime

import pytest
from resources import expected_data
from swesarr_pytools.access_swesarr import AccessSwesarr


@pytest.fixture
def access_swesarr():
    return AccessSwesarr()


def test_retrieve_meta(access_swesarr):
    assert access_swesarr.data_meta == expected_data.EXPECTED_META
    assert access_swesarr.flight_names == expected_data.EXPECTED_FOLDER_NAMES
    assert access_swesarr.flight_dates == expected_data.EXPECTED_FOLDER_DATES


def test_validate_date(access_swesarr):
    expected_date = datetime.date(year=1995, month=10, day=5)
    assert access_swesarr.validate_date(some_date="10/5/95", date_name="This Date") == expected_date
    assert access_swesarr.validate_date(some_date="10/5/1995", date_name="This Date") == expected_date
    assert access_swesarr.validate_date(
        some_date=datetime.date(year=1995, month=10, day=5), date_name="This Date") == expected_date


def test_available_date_within_range(access_swesarr):
    retrieved_dates = access_swesarr.available_date_within_range(start_date=datetime.date(2019, 1, 1),
                                                                 end_date=datetime.date(2019, 12, 31))
    assert retrieved_dates == expected_data.EXPECTED_DATES_RANGE

    retrieved_empty_dates = access_swesarr.available_date_within_range(
        start_date=datetime.date(2018, 1, 1),
        end_date=datetime.date(2018, 12, 31)
    )
    assert retrieved_empty_dates is None



import logging
import os


def get_logger(file_name):
    """
    :param file_name: Name file
    :return: Logger object
    """
    logger_name = os.path.basename(file_name).split('.')[0]
    logger = logging.getLogger(logger_name.upper())
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(name)s - %(asctime)s - %(levelname)s - %(message)s')
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)

    if not logger.hasHandlers():
        logger.addHandler(stream_handler)

    return logger



"""
This main class has the following functionality
1. Data retrieval for SWESARR flight from https://glihtdata.gsfc.nasa.gov/files/radar/SWESARR/prerelease
2. Metadata retrieval for SWESARR flight path
3. Singular Flight path download for one flight line but six frequencies
4. Singular frequency download for a given flight line
5. Flight path search using date
"""

__author__ = "Evi Ofekeze"
__authors__ = ["HP Marshal"]
__contact__ = "eviofekeze@u.boisestate.edu"
__copyright__ = "Copyright 2024, Boise State University, Boise ID"
__group__ = "Cryosphere GeoPhysics and Remote Sensing Laboratory"
__credits__ = ["Evi Ofekeze", "HP Marshal"]
__email__ = "eviofekeze@u.boisestate.edu"
__maintainer__ = "developer"
__status__ = "Research"

import os
from pathlib import Path
from typing import Dict, List, Union
from dataclasses import dataclass, field

import requests
from urllib.parse import urljoin
from bs4 import BeautifulSoup

from datetime import datetime
from datetime import date
from utils.swesarr_utils import get_logger

logger = get_logger(__file__)


@dataclass
class AccessSwesarr:
    url: str = f"https://glihtdata.gsfc.nasa.gov/files/radar/SWESARR/prerelease"
    ext: str = ""
    flight_names: List = field(init=False)
    flight_dates: List = field(init=False)
    data_meta: Dict = field(init=False)

    def __post_init__(self):
        self.retrieve_meta()

    def retrieve_meta(self) -> None:
        response = requests.get(self.url)
        if response.ok:
            response_text = response.text
        else:
            return response.raise_for_status()

        soup = BeautifulSoup(response_text, features='html.parser')
        self.flight_names = [f"{self.url}{node.get('href')}" for node in soup.find_all('a') if
                             node.get('href').endswith(self.ext)]
        self.flight_names = self.flight_names[5:-1]
        self.flight_names = [name[62:] for name in self.flight_names]
        self.flight_dates = [datetime.strptime(date_.split('_')[4], '%y%m%d').date() for date_ in self.flight_names]
        self.data_meta = dict(zip(self.flight_names, self.flight_dates))
        return

    @staticmethod
    def validate_date(some_date: Union[str, date], date_name: str) -> date:
        """
        Parses date and returns valid data

        :param some_date: str | datetime.date
            date to be validated
        :param date_name: Name for error logging
        :return:
        """
        date_formats = ["%m/%d/%Y", "%m/%d/%y", "%Y-%m-%d"]

        if isinstance(some_date, date):
            return some_date

        for date_format in date_formats:
            try:
                parsed_some_date = datetime.strptime(some_date, date_format).date()
                return parsed_some_date
            except (ValueError, TypeError):
                continue
        raise ValueError(f"{date_name} format not recognised")

    def available_date_within_range(self, start_date: Union[str, date], end_date: Union[str, date]) -> List:
        """
        Search for swesarr flight within a time period

        :param start_date: The start date range
        :param end_date:  The end date range
        :return: List of available dates
        """
        start_date = self.validate_date(start_date, date_name="Start Date")
        end_date = self.validate_date(end_date, date_name="End Date")

        result = [date_ for date_ in self.flight_dates if (start_date <= date_ <= end_date)]
        if not result:
            logger.info(f"No flight path was found within the specified range")
        else:
            return result

    @staticmethod
    def download_tif(base_url: str, file_name: str) -> None:
        """
        Helper function to download a specific fly
        :param base_url: The root Url
        :param file_name: The file name corresponding the band
        :return: None
        """
        response = requests.get(base_url)
        with open(file_name, 'wb') as file:
            for chunk in response.iter_content(chunk_size=5 * 1024 * 1024):
                if chunk:
                    file.write(chunk)
        logger.info(f"Downloaded: {base_url}")
        return

    def download_data(self, flight_path: str, folder_name: str, band: str = "all", dem: bool = False) -> None:
        """
        Downloads data for a specified flight path.

        This method downloads the associated data into the specified folder. It also
        allows for filtering by band and includes an option to download DEM (Digital
        Elevation Model) data.

        :param flight_path: str
            The name of the flight path to be downloaded
        :param folder_name: str
            The name of the folder where the downloaded data will be stored
        :param band: str, optional
            Specifies the band to download (e.g., "X", "KuLo", "KuHi"). Defaults to "all",
            which downloads all available bands
        :param dem: bool, optional
            If True, the Digital Elevation Model (DEM) data is downloaded along with the
            band data. Defaults to False
        :return: None
        """
        folder_name = Path(folder_name)
        band = band.lower()
        complete_url = f"{self.url}/{flight_path}/"
        response = requests.get(complete_url)
        if response.status_code != 200:
            print(f"Failed to fetch page content: {response.status_code}")
            return
        soup = BeautifulSoup(response.content, features='html.parser')

        tif_links = [urljoin(complete_url, link['href']) for link in soup.find_all(name='a', href=True) if
                     link['href'].endswith('.tif')]
        if dem:
            tif_links = tif_links
        else:
            tif_links[:] = [item for item in tif_links if "dem" not in item]

        fully_qualified_folder_name = Path(f"{folder_name}/{flight_path}")
        os.makedirs(fully_qualified_folder_name, exist_ok=True)
        if band == "all":
            for link in tif_links:
                file_name = os.path.join(fully_qualified_folder_name, os.path.basename(link))
                self.download_tif(link, file_name)

        elif band == "x":
            tif_links = [item for item in tif_links if "09225" in item]
            for link in tif_links:
                file_name = os.path.join(fully_qualified_folder_name, os.path.basename(link))
                self.download_tif(link, file_name)

        elif band == "kulo":
            tif_links = [item for item in tif_links if "13225" in item]
            for link in tif_links:
                file_name = os.path.join(fully_qualified_folder_name, os.path.basename(link))
                self.download_tif(link, file_name)

        elif band == "kuhi":
            tif_links = [item for item in tif_links if "17225" in item]
            for link in tif_links:
                file_name = os.path.join(fully_qualified_folder_name, os.path.basename(link))
                self.download_tif(link, file_name)
        return

    def download_all_data(self, folder_name: str, band: str = "all", dem: bool = False) -> None:
        """
        Downloads data for all flights.

        This method iterates over all flight paths generated in `self.flight_names` list
        and downloads the associated data into the specified folder. It also
        allows for filtering by band and includes an option to download DEM (Digital
        Elevation Model) data.

        :param folder_name: str
            The name of the folder where the downloaded data will be stored
        :param band: str, optional
            Specifies the band to download (e.g., "X", "KuLo", "KuHi"). Defaults to "all",
            which downloads all available bands
        :param dem: bool, optional
            If True, the Digital Elevation Model (DEM) data is downloaded along with the
            band data. Defaults to False

        :return: None
        """
        for this_flight_path in self.flight_names:
            self.download_data(
                flight_path=this_flight_path,
                folder_name=folder_name,
                dem=dem,
                band=band
            )
        return


if __name__ == '__main__':
    logger.info(f"Nothing to report... Moving on")



import datetime

import pytest
from swesarr_pytools.data_tools import ReadSwesarr


@pytest.fixture
def read_swesarr_v1():
    return ReadSwesarr(flight_path="/opt/project/resources/swesarr_data/GRMCT1_13801_19005_010_191106_225_XX_01",
                       version="v1")

@pytest.fixture
def read_swesarr_v3():
    return ReadSwesarr(flight_path="/opt/project/resources/swesarr_data_v3/19004_011",
                       version="v3")


def test_get_data_files(read_swesarr_v1 , read_swesarr_v3):
    expected_list_v1 = ['GRMCT1_13801_19005_010_191106_13225VH_XX_01.tif',
                        'GRMCT1_13801_19005_010_191106_17225VH_XX_01.tif',
                        'GRMCT1_13801_19005_010_191106_09225VV_XX_01.tif',
                        'GRMCT1_13801_19005_010_191106_17225VV_XX_01.tif',
                        'GRMCT1_13801_19005_010_191106_13225VV_XX_01.tif',
                        'GRMCT1_13801_19005_010_191106_09225VH_XX_01.tif']

    expected_list_v3 = ['SWESARR_northdescend_BasicOps_SWProgXcvr_Box2_2019Nov05-15.31.58.297411_2019Nov05-15.33.36'
                        '.745263_XVV_MC_neg.tif',
                        'SWESARR_northdescend_BasicOps_SWProgXcvr_Box2_2019Nov05-15.31.58.297411_2019Nov05-15.33.36'
                        '.745263_KlVV_MC_neg.tif',
                        'SWESARR_northdescend_BasicOps_SWProgXcvr_Box2_2019Nov05-15.31.58.297411_2019Nov05-15.33.36'
                        '.745263_XVVINC_MC_neg.mli.inc_deg.tif',
                        'SWESARR_northdescend_BasicOps_SWProgXcvr_Box2_2019Nov05-15.31.58.297411_2019Nov05-15.33.36'
                        '.745263_XVH_MC_neg.tif',
                        'SWESARR_northdescend_BasicOps_SWProgXcvr_Box2_2019Nov05-15.31.58.297411_2019Nov05-15.33.36'
                        '.745263_KhVH_MC_neg.tif',
                        'SWESARR_northdescend_BasicOps_SWProgXcvr_Box2_2019Nov05-15.31.58.297411_2019Nov05-15.33.36'
                        '.745263_KlVH_MC_neg.tif',
                        'SWESARR_northdescend_BasicOps_SWProgXcvr_Box2_2019Nov05-15.31.58.297411_2019Nov05-15.33.36'
                        '.745263_KhVV_MC_neg.tif',
                        'SWESARR_northdescend_BasicOps_SWProgXcvr_Box2_2019Nov05-15.31.58.297411_2019Nov05-15.33.36'
                        '.745263_XVVOFNA_MC_neg.mli.offnadir_deg.tif']

    assert read_swesarr_v1.get_data_files() == expected_list_v1
    assert read_swesarr_v3.get_data_files() == expected_list_v3


